##############################
# This code implements some of the (corrected) expressions from
#   A.J. Baddeley, R.A. Moyeed, C.V.Howard, A.Boyde,
#   Appl. Statist., 42, 641 (1993).
#
# The implementation follows closely the C-code in
# sphefrac.c (written by A.J. Baddeley) from the R package
# spatstat https://spatstat.org/ .
# In case all errors are due to M. Kerscher who reimplemented 
# the functions in python. See also
#    M. Kerscher, submitted to A&A (March 24, 2022)
#    https://arxiv.org/abs/2203.13288

import numpy as np

# C divided by pi
#   assumes a,b,c > 0
def c3(a,b,c):
    if a*a +b*b + c*c >= 1.0:
        return 0.

    za = np.sqrt(1. - b*b - c*c)
    zb = np.sqrt(1. - a*a - c*c)
    zc = np.sqrt(1. - a*a - b*b)

    sum = np.arctan2(zb, a*c) + np.arctan2(za, b*c) + np.arctan2(zc, a*b)
    sum += -a*np.arctan2(zb, c) + a*np.arctan2(b, zc) 
    sum += -b*np.arctan2(za, c) + b*np.arctan2(a, zc) 
    sum += -c*np.arctan2(zb, a) + c*np.arctan2(b, za)

    return sum/np.pi - 1.


# C(a,b,0) divided by pi
# assumes a,b, > 0
def c2(a,b):
    z2 = 1. - a*a - b*b
    if z2 < 0:
        return 0.
    z = np.sqrt(z2)
    s = np.arctan2(z,a*b) - a*np.arctan2(z,b) - b*np.arctan2(z,a)
    return s/np.pi


# this is A1 divided by 4 pi r^2
def a1(t,r):
    if t>=r:
        return 0.
    return (1. - t/r)/2.


# this is A2 divided by 4 pi r^2
def a2(t1,t2,r):
    return c2(t1/r, t2/r)/2.


# this is A3 divided by 4 pi r^2
def a3(t1,t2,t3,r):
    return c3(t1/r, t2/r, t3/r)/4.


# Baddeley et al, 1993, analog to eq.(27)
def sphesfrac(x, r,L):
    p = x  # left lower [0 0 0]
    q = L - x
    
    sum=0.0
    for i in range(3):
        sum += a1(p[i],r) + a1(q[i],r)

    for i in range(3):
        for j in range(i+1,3):
            sum -= a2(p[i],p[j],r) + a2(p[i],q[j],r)
            sum -= a2(q[i],p[j],r) + a2(q[i],q[j],r)

    sum += a3(p[0],p[1],p[2],r)
    sum += a3(p[0],p[1],q[2],r) + a3(p[0],q[1],p[2],r) + a3(p[0],q[1],q[2],r)
    sum += a3(q[0],p[1],p[2],r) + a3(q[0],p[1],q[2],r) + a3(q[0],q[1],p[2],r)
    sum += a3(q[0],q[1],q[2],r)

    return 1.0-sum


# the fraction of the area inside
def area_fraction(x, r,L):
    return 4*np.pi*r*r * sphesfrac(x, r,L)
