//  routines for calculating the two-point correlation function
//  as described in
//    M. Kerscher, submitted to A&A (March 24, 2022)
//    https://arxiv.org/abs/2203.13288

#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/numpy.h>

#include <cmath>
#include <vector>
#include <algorithm>

using namespace std;
namespace py = pybind11;

/////////////////////////////////////////////////

// define/overload the reduction operator + for  vector<int> 
//  transform: every call to + reduction iterates through the whole vector using transform
//  initializer: all private variables/vectors are initialized with 0
#pragma omp declare reduction( + : vector<long long> : \
  transform( omp_in.begin(),omp_in.end(),omp_out.begin(),omp_out.begin(), plus<long long>() )) \
  initializer(omp_priv(omp_orig.size(),0))


// Euclidean distance
double dist(double x1, double y1, double z1,
	    double x2, double y2, double z2) {
  return sqrt(pow(x1-x2,2.)+pow(y1-y2,2.)+pow(z1-z2,2.));
}


// x,y,z in one matrix dat, row major
//  x_1 y_1 z_1
//  .....
// no double counting
void paircount(double* dat, size_t nrow, size_t ncol,
	       vector<double>& r, vector<long long>& npairs ) {

  // reset result vector
  for(size_t i=0; i<npairs.size(); ++i){
    npairs[i] = 0;
  }

  double xi,yi,zi;
  // loop over points i
  for(size_t i=0; i<nrow; ++i) {
    xi = dat[i*ncol + 0];
    yi = dat[i*ncol + 1];
    zi = dat[i*ncol + 2];

    // reduction: vector  npairs  are added in the threads 
    //    see overload/declare + for vectors above
#pragma omp parallel for reduction(+:npairs)
    // loop over other points j (no double counting)
    for(size_t j=i+1; j<nrow; ++j) {
      double xj = dat[j*ncol + 0];
      double yj = dat[j*ncol + 1];
      double zj = dat[j*ncol + 2];
      double d = dist(xi,yi,zi, xj,yj,zj);
      // search and add up, discard if d too large or too small 
      if (d < r.back() &&  d > r.front()){
	// binary search in the vector of radii r
	vector<double>::iterator it = lower_bound(r.begin(), r.end(), d);
	int irad = it - r.begin() - 1; // index = position - 1

	npairs[irad] += 1;
      }
    } // end loop over other points j
    // end omp parallel
    
  } // end loop over points i

  return;
}


// x,y,z in one matrix for each of dat1 and dat2 in row major
//  x_1 y_1 z_1
//  .....
void paircount2(double* dat1, size_t nrow1, double* dat2, size_t nrow2, size_t ncol,
	       vector<double>& r, vector<long long>& npairs ) {

  // reset result vector
  for(size_t i=0; i<npairs.size(); ++i){
    npairs[i] = 0;
  }

  double xi,yi,zi;
  // loop over points i
  for(size_t i=0; i<nrow1; ++i) {
    xi = dat1[i*ncol + 0];
    yi = dat1[i*ncol + 1];
    zi = dat1[i*ncol + 2];

    // reduction: vector  npairs  are added in the threads 
    //    see overload/declare + for vectors above
#pragma omp parallel for reduction(+:npairs)
    // loop over other points j
    for(size_t j=0; j<nrow2; ++j) {
      double xj = dat2[j*ncol + 0];
      double yj = dat2[j*ncol + 1];
      double zj = dat2[j*ncol + 2];
      double d = dist(xi,yi,zi, xj,yj,zj);
      // search and add up, discard if d too large or too small 
      if (d < r.back() &&  d > r.front()){
	// binary search in the vector of radii r
	vector<double>::iterator it = lower_bound(r.begin(), r.end(), d);
	int irad = it - r.begin() - 1; // index = position - 1

	npairs[irad] += 1;
      }
    } // end loop over other points j
    // end omp parallel
    
  } // end loop over points i

  return;
}


///////////////////////////////////////////////////////////////////////////////////////


// wrap C++ function with NumPy array IO
//  dat is accessed via pointer to numpy storage
//  rad is copied from numpy to vector
//  res is returned (copied) into a numpy array 
py::array_t<double> py_paircount( py::array_t<double, py::array::c_style | py::array::forcecast> dat,
				  py::array_t<double, py::array::c_style | py::array::forcecast> rad )
{
  // get the pointer to the internal storage of the ndarray dat
  //   do not copy it, it might be large
  // see http://people.duke.edu/~ccc14/cspy/18G_C++_Python_pybind11.html
  // and https://docs.python.org/3/c-api/buffer.html
  py::buffer_info datinfo = dat.request();
  auto datptr = static_cast<double *>(datinfo.ptr);
  int ndim = datinfo.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow = datinfo.shape[0];
  size_t ncol = datinfo.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z

  // copy rad to vector radii (small enough)
  py::buffer_info radinfo = rad.request();
  auto radptr = static_cast<double *>(radinfo.ptr);
  ndim = radinfo.ndim;
  assert( ndim==1 );
  size_t nrad = radinfo.shape[0];
  vector<double> radii(nrad);
  for(size_t i=0; i<nrad; ++i) {
    radii[i] = radptr[i];
  }

  // results are stored in this vector
  size_t nint = radii.size()-1;
  vector<long long> npairs(nint,0);

  // call function
  //   datptr is in row major order, the others are vectors
  paircount(datptr,nrow,ncol, radii,npairs);

  // create new numpy array and copy results into
  // https://stackoverflow.com/questions/44659924/returning-numpy-arrays-via-pybind11
  // https://github.com/pybind/pybind11/issues/1429
  py::array_t<double> res(nint); // ndarray nint   
  auto pc = res.mutable_unchecked<1>();  // 1d ndarray
  for(size_t i=0; i<nint; ++i) {
    pc(i) = double(npairs[i]);
  }
  
  return res;
}

////////////////////////////////////////////

// wrap C++ function with NumPy array IO
//  dat1 is accessed via pointer to numpy storage
//  dat2 is accessed via pointer to numpy storage
//  rad is copied from numpy to vector
//  res is returned (copied) into a numpy array 
py::array_t<double> py_paircount2( py::array_t<double, py::array::c_style | py::array::forcecast> dat1,
				   py::array_t<double, py::array::c_style | py::array::forcecast> dat2,
				   py::array_t<double, py::array::c_style | py::array::forcecast> rad )
{
  // get the pointer to the internal storage of the ndarray dat
  //   do not copy, it might be large
  // see http://people.duke.edu/~ccc14/cspy/18G_C++_Python_pybind11.html
  // and https://docs.python.org/3/c-api/buffer.html
  //
  // dat1
  py::buffer_info dat1info = dat1.request();
  auto dat1ptr = static_cast<double *>(dat1info.ptr);
  int ndim = dat1info.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow1 = dat1info.shape[0];
  size_t ncol = dat1info.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z
  // dat2
  py::buffer_info dat2info = dat2.request();
  auto dat2ptr = static_cast<double *>(dat2info.ptr);
  ndim = dat2info.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow2 = dat2info.shape[0];
  ncol = dat2info.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z

  // copy rad to vector radii (small enough)
  py::buffer_info radinfo = rad.request();
  auto radptr = static_cast<double *>(radinfo.ptr);
  ndim = radinfo.ndim;
  assert( ndim==1 );
  size_t nrad = radinfo.shape[0];
  vector<double> radii(nrad);
  for(size_t i=0; i<nrad; ++i) {
    radii[i] = radptr[i];
  }

  // results are stored in this vector
  size_t nint = radii.size()-1;
  vector<long long> npairs(nint,0);

  // call function
  //   datptr is in row major order, the others are vectors
  paircount2(dat1ptr,nrow1, dat2ptr,nrow2, ncol, radii,npairs);

  // create new numpy array and copy results into
  // https://stackoverflow.com/questions/44659924/returning-numpy-arrays-via-pybind11
  // https://github.com/pybind/pybind11/issues/1429
  py::array_t<double> res(nint); // ndarray nint   
  auto pc = res.mutable_unchecked<1>();  // 1d ndarray
  for(size_t i=0; i<nint; ++i) {
    pc(i) = double(npairs[i]);
  }
  
  return res;
}

/////////////////////////////////////////////////
    

// count the number of points inside [0,L0]x[0,L1]x[0,L2]
long n_in_box(double* dat, size_t nrow, size_t ncol, vector<double>& L)
{
  // we count
  long inwin=0;
  double x,y,z;
  size_t i;
  
  // loop over points i
#pragma omp parallel for private(x,y,z,i) shared(nrow,ncol,dat,L) reduction(+:inwin)
  for(i=0; i<nrow; ++i) {
    x = dat[i*ncol + 0];
    y = dat[i*ncol + 1];
    z = dat[i*ncol + 2];
    if(x>0 && x<L[0] && y>0 && y<L[1] && z>0 && z<L[2]) {
      inwin++;
    }
  }
  return inwin;
}


// wrap C++ function with NumPy array IO
//  dat is accessed via pointer to numpy storage
//  L is copied from numpy to vector
//  res is returned (copied) into a numpy array 
long py_ninbox( py::array_t<double, py::array::c_style | py::array::forcecast> dat,
		py::array_t<double, py::array::c_style | py::array::forcecast> Lin )
{
  // get the pointer to the internal storage of the ndarray dat
  //   do not copy, it might be large
  // see http://people.duke.edu/~ccc14/cspy/18G_C++_Python_pybind11.html
  // and https://docs.python.org/3/c-api/buffer.html
  py::buffer_info datinfo = dat.request();
  auto datptr = static_cast<double *>(datinfo.ptr);
  int ndim = datinfo.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow = datinfo.shape[0];
  size_t ncol = datinfo.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z

  // copy box size Lin to vector L (small enough)
  py::buffer_info Lininfo = Lin.request();
  auto Linptr = static_cast<double *>(Lininfo.ptr);
  ndim = Lininfo.ndim;
  assert( ndim==1 );
  size_t nL = Lininfo.shape[0];
  assert( nL==3 );
  vector<double> L(nL);
  for(size_t i=0; i<nL; ++i) {
    L[i] = Linptr[i];
  }

  // call function
  //   datptr is in row major order, the L is a vector
  long nin = n_in_box(datptr,nrow,ncol, L);
  
  return nin;
}


/////////////////////////////////////////////////


// count the number of point pairs with distance in radius bins
void count_pairs_in_radius_range(double* dat1, double* dat2, size_t nrow, size_t ncol,
				 vector<double>& r, vector<long long>& ninrad )
{
  // we assume assert( nrow1==nrow2 ); assert( ncol==3 );
  double x1,y1,z1, x2,y2,z2, d;
  size_t i;
  
  // loop over points i
#pragma omp parallel for private(x1,y1,z1,x2,y2,z2,d,i) shared(nrow,ncol,dat1,dat2,r) reduction(+:ninrad)
  for(i=0; i<nrow; ++i) {
    x1 = dat1[i*ncol + 0];
    y1 = dat1[i*ncol + 1];
    z1 = dat1[i*ncol + 2];
    x2 = dat2[i*ncol + 0];
    y2 = dat2[i*ncol + 1];
    z2 = dat2[i*ncol + 2];
    d = dist(x1,y1,z1, x2,y2,z2);
    // search and add up, discard if d too large or too small 
    if (d < r.back() &&  d > r.front()){
      // binary search in the vector of radii r
      vector<double>::iterator it = lower_bound(r.begin(), r.end(), d);
      int irad = it - r.begin() - 1; // index = position - 1
      ninrad[irad] += 1;
    }
  }
  return;
}


// wrap C++ function with NumPy array IO
//  dat1 is accessed via pointer to numpy storage
//  dat2 is accessed via pointer to numpy storage
//  rad is copied from numpy to vector
//  res is returned (copied) into a numpy array 
py::array_t<double> py_fninradrange( py::array_t<double, py::array::c_style | py::array::forcecast> dat1,
				    py::array_t<double, py::array::c_style | py::array::forcecast> dat2,
				    py::array_t<double, py::array::c_style | py::array::forcecast> rad )
{
  // get the pointer to the internal storage of the ndarray dat
  //   do not copy, it might be large
  // see http://people.duke.edu/~ccc14/cspy/18G_C++_Python_pybind11.html
  // and https://docs.python.org/3/c-api/buffer.html
  //
  // dat1
  py::buffer_info dat1info = dat1.request();
  auto dat1ptr = static_cast<double *>(dat1info.ptr);
  int ndim = dat1info.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow1 = dat1info.shape[0];
  size_t ncol = dat1info.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z
  // dat2
  py::buffer_info dat2info = dat2.request();
  auto dat2ptr = static_cast<double *>(dat2info.ptr);
  ndim = dat2info.ndim;
  assert( ndim==2 );  // 2d array
  size_t nrow2 = dat2info.shape[0];
  ncol = dat2info.shape[1];
  assert( ncol==3 ); // 3 cols x,y,z

  // check same size
  assert( nrow1==nrow2 );
  
  // copy rad to vector radii (small enough)
  py::buffer_info radinfo = rad.request();
  auto radptr = static_cast<double *>(radinfo.ptr);
  ndim = radinfo.ndim;
  assert( ndim==1 );
  size_t nrad = radinfo.shape[0];
  vector<double> radii(nrad);
  for(size_t i=0; i<nrad; ++i) {
    radii[i] = radptr[i];
  }

  // results are stored in this vector
  size_t nint = radii.size()-1;
  vector<long long> ninrange(nint,0);

  // call function
  //   datptr is in row major order, the others are vectors
  count_pairs_in_radius_range(dat1ptr,dat2ptr,nrow1,ncol, radii, ninrange);
  
  // create new numpy array and copy results into
  // https://stackoverflow.com/questions/44659924/returning-numpy-arrays-via-pybind11
  // https://github.com/pybind/pybind11/issues/1429
  py::array_t<double> res(nint); // ndarray nint   
  auto pc = res.mutable_unchecked<1>();  // 1d ndarray
  for(size_t i=0; i<nint; ++i) {
    pc(i) = double(ninrange[i])/nrow1;
  }
  
  return res;
}


/////////////////////////////////////////////////

// wrap as Python module
PYBIND11_MODULE(cppauxiliary, m)
{
  m.doc() = "paircount, paircount2, ninbox, fninradrange";
  m.def("paircount", &py_paircount, "paircount, n.o. pairs, no double counting");
  m.def("paircount2", &py_paircount2, "paircount2, n.o. pairs between dat1 dat2");
  m.def("ninbox", &py_ninbox, "n.o. points inside box");
  m.def("fninradrange", &py_fninradrange, "fraction of point pairs in radius range");
}
