#  routines for calculating the two-point correlation function
#  as described in
#    M. Kerscher, submitted to A&A (March 24, 2022)
#    https://arxiv.org/abs/2203.13288

import numpy as np
from scipy.stats import qmc
import scipy.integrate as integrate

# the area fraction 
from baddeley import area_fraction
# the cpp helpers using openmp
import cppauxiliary as caux


###############
# exact

#   int_r^R \overline{\gamma}_W(s) s^2 ds
def int_isotropic_setcov_cuboid(r,R, L):
    v = 1./3. * np.prod(L) * (R**3 - r**3)
    v -= 1./8. * (L[0]*L[1] + L[0]*L[2] + L[1]*L[2]) * (R**4 - r**4)
    v += 2./(15.*np.pi) * np.sum(L) * (R**5 - r**5)
    v -= 1./(24.*np.pi) * (R**6 - r**6)
    return v


# area fraction averaged over all points
def average_area_fraction(xn, r,L):
    a = 0.
    for q in xn:
        a += area_fraction(q, r,L)
    return a/len(xn)


# the xi LS with exact DR and RR
def xi_LS_exact(pts,radii,L):
    ### DD
    DD = caux.paircount(pts, radii)
    DD = 2.0*DD/(len(pts)*len(pts))

    ### RR (without randoms)
    # calculate the reference isotropic set covariance
    intisc = np.zeros(len(radii)-1)
    for i in range(len(radii)-1):
        r = radii[i]
        R = radii[i+1]
        intisc[i] = int_isotropic_setcov_cuboid(r,R, L) 
    # RR = 4 pi / |w|^2 * avisc
    RR = intisc * 4.*np.pi/(np.prod(L)**2)
    
    ### DR (without randoms)
    avafrac = np.zeros(len(radii)-1)
    for i in range(len(radii)-1):
        r = radii[i]
        R = radii[i+1]
        resquad = integrate.quad(lambda q: average_area_fraction(pts,q,L), r,R)
        # for higher accuracy
        #resquad = integrate.quad(lambda q: average_area_fraction(pts,q,L), 
        #                         r,R, epsabs=0., epsrel=1.e-10)
        avafrac[i] = resquad[0]
    # DR = average_area_fraction / |W|
    DR = avafrac / np.prod(L)
    
    # combine and return
    return (DD-2.*DR)/RR + 1.


##################################
# standard

def xi_LS_standard(pts, radii, L, rns):
    # DD
    DD = caux.paircount(pts, radii)
    DD = 2.0*DD/(len(pts)*len(pts))
    # RR 
    RR = caux.paircount(rns, radii)
    RR = 2.0*RR/(len(rns)*len(rns))
    # DR 
    DR = caux.paircount2(pts, rns, radii)
    DR = DR/(len(pts)*len(rns))
    #
    xi = (DD-2.*DR)/RR + 1.
    return xi


##################################
# improved

def DRshell(radii, L,cldat, rn3d):
    DRsh = np.zeros(len(radii)-1)
    # randoms on unit sphere
    #    (from the first two dimensions of rn3d)
    phi = 2.*np.pi* (rn3d.T)[0]
    theta = np.arccos(2.*(rn3d.T)[1] - 1.)
    s = np.sin(theta)
    # we need [ [..x..],[..y..],[..z..] ] --- see below rnshell
    unitsphere = np.vstack((np.cos(phi)*s, np.sin(phi)*s, np.cos(theta)))
    # radius loop
    for i in range(len(radii)-1):
        r = radii[i]
        R = radii[i+1]
        # random radius in shell [r,R]
        #    (from the third dimension of rn3d)
        t = (np.power(R,3.) - np.power(r,3.))*(rn3d.T)[2] + np.power(r,3.)
        u = np.power(t, 1./3.)
        # random points in shell with [r,R]
        rnshell = (u*unitsphere).T
        # calculate avrage n.o. points inside W
        avs = 0.
        for pt in cldat:
            y = rnshell + pt
            avs += caux.ninbox(y,L)/len(y)
        avs /= len(cldat)
        DRsh[i] = 4.*np.pi/3.*(R**3-r**3)/np.product(L) * avs
    return DRsh


#  rn3d: quasi random points in [0,1]^3
#  rn6d: quasi random points in [0,1]^6
def xi_LS_rqmc(pts,radii,L,  rn3d, rn6d):
    ### DD
    DD = caux.paircount(pts, radii)
    DD = 2.0*DD/(len(pts)*len(pts))
    
    ### DR_shell  via points in shell
    DR = DRshell(radii, L,pts, rn3d)
    
    ### RR_2x3d   via 2x3d paircount
    qrn1 = rn6d[:,:3]*L
    qrn2 = rn6d[:,3:]*L
    RR = caux.paircount2(qrn1, qrn2, radii)
    RR = RR/(len(qrn1)*len(qrn2))
        
    # combine
    return (DD-2.*DR)/RR + 1.
