#  an example calculation for the methods discussed in
#    M. Kerscher, submitted to A&A (March 24, 2022)
#    https://arxiv.org/abs/2203.13288

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import qmc

# the python helpers
import pyauxiliary as paux

# global random number generator
rng = np.random.default_rng(12345)
# quasi random sampler
qmcsampler3d = qmc.Halton(d=3)
qmcsampler6d = qmc.Halton(d=6)

# read the points
pts = np.loadtxt("points.pt")
# the box [0,1]^3
L = np.array([1.,1.,1.])
# build a list with radius intervalls
r = np.linspace(0,0.3,num=16)


### xi LS with exact geometric RR and DR
xiLSexact = paux.xi_LS_exact(pts,r,L)
print("# exact LS done")

### standard xi LS from paircounts
nr=1000
rn = rng.uniform(size=(nr,3))
xiLS = paux.xi_LS_standard(pts,r,L, rn)
print("# standard LS done")

### xi LS with rqmc
# 6d points from [0,1]^6
n6d = 1000
rn6d = qmcsampler6d.random(n6d)
# 3d points from [0,1]^3
n3d = 1000
rn3d = qmcsampler3d.random(n3d)
xiLSrqmc = paux.xi_LS_rqmc(pts,r,L, rn3d,rn6d)
print("# rqmc LS done")


### output
print(f"#  {len(pts)} points in a box")
print(f"#  standard LS, Nr={nr}")
print(f"#  rqmc LS, n3d={n3d}, n6d={n6d}")
print("#  rv  r  R  xiLSexact xiLS xiLSrqmc")

rv = np.zeros(len(xiLS))
for i in range(len(xiLS)):
    rl = r[i]
    ru = r[i+1]
    # volume weighted radius
    rv[i] = np.power((np.power(ru,3.) + np.power(rl,3.))/2, 1./3.)
    #
    print(f"{rv[i]:.3f} {rl:.3f} {ru:.3f} ", end="")
    print(f"{xiLSexact[i]:e} {xiLS[i]:e} {xiLSrqmc[i]:e}")


### plot
fig, ax = plt.subplots()
ax.plot(rv,xiLSexact, rv,xiLS, rv,xiLSrqmc)
ax.set_ylim((-0.05,0.2))
ax.legend(["LS exact", "LS standard", "LS rqmc"])
fig.savefig("xi.pdf")
#
fig, ax = plt.subplots()
ax.loglog(rv,xiLSexact, rv,xiLS, rv,xiLSrqmc)
ax.set_ylim((0.001,10))
ax.legend(["LS exact", "LS standard", "LS rqmc"])
fig.savefig("xilog.pdf")
